#include "msr-index.h"
#include "emu.h"
#include "emu-main.c"

/* --------------------------------------------------------------------- */
/* helper functions                                                      */

int bounce_trap(struct regs_64 *regs, int trapno, int cbno)
{
    uint64_t *stack, rip, rsp, stack_cs, stack_rflags;
    int error_code = 0;
    int interrupt = 0;
    int failsafe = 0;
    int k = 0;

    if (trapno >= 0) {
	/* trap bounce */
	rip  = xen_ctxt.trap_ctxt[trapno].address;
	if (TI_GET_IF(&xen_ctxt.trap_ctxt[trapno]))
	    interrupt = 1;
	if (trapno < sizeof(trapinfo)/sizeof(trapinfo[0]))
	    error_code = trapinfo[trapno].ec;
	if (14 == trapno) /* page fault */
	    shared_info.vcpu_info[0].arch.cr2 = read_cr2();
    }
    if (cbno >= 0) {
	/* callback */
	rip  = xencb[cbno];
	switch (cbno) {
	case CALLBACKTYPE_event:
 	    interrupt = 1;
	    break;
	}
    }

    if (!rip)
	panic("no guest trap handler");

    /* set interrupt flag depending on event channel mask */
    stack_rflags = regs->rflags & ~X86_EFLAGS_IF;
    if (!shared_info.vcpu_info[0].evtchn_upcall_mask)
	stack_rflags |= X86_EFLAGS_IF;

    /* old evtchn_upcall_mask is saved in cs slot on the stack */
    stack_cs = regs->cs | ((uint64_t)shared_info.vcpu_info[0].evtchn_upcall_mask << 32);
    if (interrupt)
	shared_info.vcpu_info[0].evtchn_upcall_mask = 1;

    if (!is_kernel()) {
	/* user mode */
	switch_mode();
	rsp = xen_ctxt.kernel_sp;
    } else {
	/* kernel mode */
	stack_cs &= ~3;         /* signal kernel mode */
	rsp = regs->rsp & ~0xf; /* align stack */
    }
    stack = (void*)(rsp);

#if 0
    if ((stack_cs & 3)  ||  cbno == CALLBACKTYPE_syscall) /* from user mode */
	printk("%s: code %" PRIx64 " stack %" PRIx64 "\n",
	       __FUNCTION__, rip, rsp);
#endif
    
    stack[-(++k)] = regs->ss;        // push ss
    stack[-(++k)] = regs->rsp;       // push rsp
    stack[-(++k)] = stack_rflags;    // push rflags
    stack[-(++k)] = stack_cs;        // push cs
    stack[-(++k)] = regs->rip;       // push rip
    if (error_code)
	stack[-(++k)] = regs->error; // push error code

    if (failsafe) {
	/* push segment registers */;
    }

    stack[-(++k)] = regs->r11;       // push r11
    stack[-(++k)] = regs->rcx;       // push rcx

    /* prepare emu stack, so iret jumps to the kernels handler. */
    regs->rip     = rip;
    regs->cs      = FLAT_KERNEL_CS;
    regs->rflags &= EFLAGS_TRAPMASK;
    regs->rsp     = rsp - 8*k;
    regs->ss      = FLAT_KERNEL_SS;

#if 0
    if (cbno == CALLBACKTYPE_syscall)
	regs->rflags |= X86_EFLAGS_TF;
#endif

    return 0;
}

/* --------------------------------------------------------------------- */

static void tss_init(int vcpu)
{
    uint64_t base;
    int size, idx = tss(vcpu);

    xen_tss.rsp0 = (intptr_t)stack_high;

    base = (intptr_t)(&xen_tss);
    size = sizeof(xen_tss)-1;
    xen_gdt[ idx +0 ] = mkdesc32(base & 0xffffffff, size, 0x89, 0);
    xen_gdt[ idx +1 ].a = base >> 32;
    xen_gdt[ idx +1 ].b = 0;
    ltr(idx << 3);
}

static void idt_init(void)
{
    struct {
	uint16_t len;
	uint64_t ptr;
    } __attribute__((packed)) idtp = {
	.len = sizeof(xen_idt)-1,
	.ptr = (intptr_t)xen_idt,
    };
    int i;

    for (i = 0; i < 256; i++)
	xen_idt[i] = mkgate64(0xe008, (intptr_t)int_unknown, 0x8e);
    
    xen_idt[    0 ] = mkgate64(0xe008, (intptr_t)division_by_zero,    0x8e);
    xen_idt[    1 ] = mkgate64(0xe008, (intptr_t)debug_int1,          0x8e);
    xen_idt[    2 ] = mkgate64(0xe008, (intptr_t)nmi,                 0x8e);
    xen_idt[    3 ] = mkgate64(0xe008, (intptr_t)debug_int3,          0xee);
    xen_idt[    4 ] = mkgate64(0xe008, (intptr_t)overflow,            0xee);
    xen_idt[    5 ] = mkgate64(0xe008, (intptr_t)bound_check,         0x8e);
    xen_idt[    6 ] = mkgate64(0xe008, (intptr_t)illegal_instruction, 0x8e);
    xen_idt[    7 ] = mkgate64(0xe008, (intptr_t)no_device,           0x8e);
    xen_idt[    8 ] = mkgate64(0xe008, (intptr_t)double_fault,        0x8e);
    xen_idt[    9 ] = mkgate64(0xe008, (intptr_t)coprocessor,         0x8e);
    xen_idt[   10 ] = mkgate64(0xe008, (intptr_t)invalid_tss,         0x8e);
    xen_idt[   11 ] = mkgate64(0xe008, (intptr_t)segment_not_present, 0x8e);
    xen_idt[   12 ] = mkgate64(0xe008, (intptr_t)stack_fault,         0x8e);
    xen_idt[   13 ] = mkgate64(0xe008, (intptr_t)general_protection,  0x8e);
    xen_idt[   14 ] = mkgate64(0xe008, (intptr_t)page_fault,          0x8e);
    xen_idt[   16 ] = mkgate64(0xe008, (intptr_t)floating_point,      0x8e);
    xen_idt[   17 ] = mkgate64(0xe008, (intptr_t)alignment,           0x8e);
    xen_idt[   18 ] = mkgate64(0xe008, (intptr_t)machine_check,       0x8e);
    xen_idt[   19 ] = mkgate64(0xe008, (intptr_t)simd_floating_point, 0x8e);

    xen_idt[ 0x20 ] = mkgate64(0xe008, (intptr_t)event_callback,      0x8e);
    xen_idt[ 0x80 ] = mkgate64(0xe008, (intptr_t)int_80,              0xee);

    asm volatile("lidt %0" : : "m" (idtp));
}

static void msrs_init(void)
{
    /* syscall setup */
    wrmsrl(MSR_STAR, (((uint64_t)0xe023 << 48) |
		      ((uint64_t)0xe008 << 32)));
    wrmsrl(MSR_LSTAR, (intptr_t)syscall);
    wrmsrl(MSR_SYSCALL_MASK, X86_EFLAGS_VM | X86_EFLAGS_RF |
	   X86_EFLAGS_NT | X86_EFLAGS_DF | X86_EFLAGS_IF | X86_EFLAGS_TF);
}

/* --------------------------------------------------------------------- */
/* called from assembler                                                 */

asmlinkage void do_boot(struct regs_64 regs)
{
    uint32_t cr3_mfn;
    int vcpu = 0;
    
    printk(0, "this is emu64 (xenner %s)\n", VERSION);
    tss_init(vcpu);
    idt_init();
    paging_init();
    msrs_init();
    cr_init();
    pv_init(vcpu);
    printk(1, "emu setup done (idt, paging, msrs, cr, pv)\n");

    m2p = (void*)XEN_M2P_64;
    vminfo.running=1;
    
    /* guest paging setup */
    cr3_mfn = xen_ctxt.ctrlreg[3] >> PAGE_SHIFT;
    update_emu_mappings_64(cr3_mfn);
    write_cr3(frame_to_addr(cr3_mfn));
    
    /* guest register setup */
    regs.rax    = xen_ctxt.user_regs.rax;
    regs.rbx    = xen_ctxt.user_regs.rbx;
    regs.rcx    = xen_ctxt.user_regs.rcx;
    regs.rdx    = xen_ctxt.user_regs.rdx;
    regs.rsi    = xen_ctxt.user_regs.rsi;
    regs.rdi    = xen_ctxt.user_regs.rdi;
    regs.rbp    = xen_ctxt.user_regs.rbp;
    regs.rip    = xen_ctxt.user_regs.rip;
    regs.cs     = xen_ctxt.user_regs.cs;
    regs.rflags = xen_ctxt.user_regs.rflags;
    regs.rsp    = xen_ctxt.user_regs.rsp;
    regs.ss     = xen_ctxt.user_regs.ss;

#if 1
    asm volatile("mov %0, %%ds;\n" :: "r" (xen_ctxt.user_regs.ds) : "memory");
    asm volatile("mov %0, %%es;\n" :: "r" (xen_ctxt.user_regs.es) : "memory");
    asm volatile("mov %0, %%fs;\n" :: "r" (xen_ctxt.user_regs.fs) : "memory");
    asm volatile("mov %0, %%gs;\n" :: "r" (xen_ctxt.user_regs.gs) : "memory");
#endif

    /* jump to kernel (via iret in asm code calling us) */
    printk(1, "booting kernel (entry %" PRIx64 ":%" PRIx64 ") ...\n",
	   regs.cs, regs.rip);
}

asmlinkage void do_page_fault(struct regs_64 regs)
{
    uint64_t cr2 = read_cr2();
    
    vminfo.faults[XEN_FAULT_PAGE_FAULT]++;

    if (context_is_user(&regs)) {
	if ((regs.error & 0x01) && addr_is_kernel(cr2)) {
	    /* present, guest kernel address */
	    /* linux vsyscall page needs this fixup */
	    if (pgtable_fixup_flag(cr2, _PAGE_USER) > 0) {
		vminfo.faults[XEN_FAULT_PAGE_FAULT_FIX_USER]++;
		return;
	    }
	}
	goto bounce;
    }

    if (context_is_emu(&regs)) {
	if (fixup_extable(&regs))
	    return;
	print_page_fault_info(0, &regs, cr2);
	panic("ring0 (emu) page fault");
    }

    /* guest kernel page fault, fixup error code */
    regs.error &= ~0x04;

    if (addr_is_user(cr2))
	goto bounce;

    if ((regs.error & 0x03) == 0x03) {
	/* present, write */
	if (pgtable_fixup_flag(cr2, _PAGE_RW) > 0) {
	    vminfo.faults[XEN_FAULT_PAGE_FAULT_FIX_RO]++;
	    return;
	}
    }

    if (regs.error & 0x01) {
	/* present */
	if (pgtable_fixup_flag(cr2, _PAGE_USER) > 0) {
	    vminfo.faults[XEN_FAULT_PAGE_FAULT_FIX_USER]++;
	    return;
	}
    }
    
    print_page_fault_info(0, &regs, cr2);
    pgtable_walk(0, cr2, 0);
    panic("Oops");

bounce:
    vminfo.faults[XEN_FAULT_PAGE_FAULT_GUEST]++;
    bounce_trap(&regs, 14, -1);
}

asmlinkage void do_event_callback(struct regs_64 regs)
{
    if (shared_info.vcpu_info[0].evtchn_upcall_mask) {
	/* FIXME */
	printk(2, "%s: IRQs off\n", __FUNCTION__);
	return;
    }
    bounce_trap(&regs, -1, CALLBACKTYPE_event);
}

asmlinkage void do_int_80(struct regs_64 regs)
{
    vminfo.faults[XEN_FAULT_INT_80]++;
    bounce_trap(&regs, 0x80, -1);
}
