#include <stdarg.h>
#include <stddef.h>
#include <inttypes.h>
#include <xen/xen.h>
#include <xen/callback.h>
#include <xen/grant_table.h>
#include <xen/version.h>
#include <xen/sched.h>
#include <xen/memory.h>

#include "shared.h"

/* attributes */
#define asmlinkage __attribute__((regparm(0)))
#define page_aligned __attribute__((aligned(4096))) __attribute__((__section__ (".pgdata")))

/* arch specific bits */
#ifdef __x86_64__
# define longmode 1
# include "emu64.h"
#else
# define longmode 0
# include "emu32.h"
#endif

/* idt entry points */
extern void division_by_zero(void);
extern void debug_int1(void);
extern void nmi(void);
extern void debug_int3(void);
extern void overflow(void);
extern void bound_check(void);
extern void illegal_instruction(void);
extern void no_device(void);
extern void double_fault(void);
extern void coprocessor(void);
extern void invalid_tss(void);
extern void segment_not_present(void);
extern void stack_fault(void);
extern void general_protection(void);
extern void page_fault(void);
extern void floating_point(void);
extern void alignment(void);
extern void machine_check(void);
extern void simd_floating_point(void);
extern void event_callback(void);
extern void int_unknown(void);

#if longmode
/* 64bit only */
extern void int_80(void);
#else
/* 32bit only */
extern void xen_hypercall(void);
#endif

/* functions */
extern int memcpy_pf(void *dest, const void *src, size_t bytes);
extern int copy32_pf(uint32_t *dest, const uint32_t *src, size_t bytes);
extern int copy64_pf(uint64_t *dest, const uint64_t *src, size_t bytes);

#define EMU_PA(vaddr)  (((intptr_t)(vaddr) - (intptr_t)_vstart))
#define EMU_MFN(vaddr) (((intptr_t)(vaddr) - (intptr_t)_vstart) >> PAGE_SHIFT)
extern char       _vstart[];
extern char       _vstop[];
extern uintptr_t  _estart[];
extern uintptr_t  _estop[];
extern char       stack_low[];
extern char       stack_high[];

/* emu-data.c */
extern int grant_frames;
extern struct grant_entry page_aligned grant_table[];
struct xenner_info vminfo;

extern struct descriptor_32 page_aligned xen_gdt[8192];
extern struct vcpu_guest_context xen_ctxt;
extern struct shared_info page_aligned shared_info;

extern xen_callback_t       xencb[8];
extern struct trap_info     xentr[256];

extern uint64_t gdt_mfn[16];
extern uint64_t emu_hcalls[XEN_HCALL_MAX];
extern uint64_t emu_faults[XEN_FAULT_MAX];

struct trapinfo {
    char *name;
    int  ec;      /* has error code  */
    int  lvl;     /* debug log level */
};
extern const struct trapinfo trapinfo[32];
extern const char *cr0_bits[32];
extern const char *cr4_bits[32];
extern const char *pg_bits[32];

/* emu-main.c */
int panic(char *message);
int bounce_trap(struct regs *regs, int trapno, int cbno);
asmlinkage void do_boot(struct regs regs);
asmlinkage void do_illegal_instruction(struct regs regs);
asmlinkage void do_general_protection(struct regs regs);
asmlinkage void do_page_fault(struct regs regs);
asmlinkage void do_event_callback(struct regs regs);
asmlinkage void do_guest_forward(struct regs regs);
asmlinkage void do_lazy_fpu(struct regs regs);

/* emu-mm.c */
void paging_init(void);
void *get_pages(int count, char *purpose);

/* emu-pv.c */
void pv_init(int vcpu);

/* emu-instr.c */
void print_bits(char *msg, uint32_t old, uint32_t new, const char *names[]);
int emulate(struct regs *regs);

/* printk.c */
int vscnprintf(char *buf, size_t size, const char *fmt, va_list args);
int snprintf(char * buf, size_t size, const char *fmt, ...);
int printk(int level, const char *fmt, ...) __attribute__((format(printf, 2, 3)));

/* inline asm bits */
static inline ureg_t read_cr0(void)
{
    ureg_t val;
    asm volatile("mov %%cr0,%0\n\t"
		 : "=r" (val));
    return val;
}

static inline ureg_t read_cr2(void)
{
    ureg_t val;
    asm volatile("mov %%cr2,%0\n\t"
		 : "=r" (val));
    return val;
}

static inline ureg_t read_cr3(void)
{
    ureg_t val;
    asm volatile("mov %%cr3,%0\n\t"
		 : "=r" (val));
    return val;
}

static inline ureg_t read_cr4(void)
{
    ureg_t val;
    asm volatile("mov %%cr4,%0\n\t"
		 : "=r" (val));
    return val;
}

static inline void write_cr0(ureg_t val)
{
    asm volatile("mov %0, %%cr0"
		 :: "r" (val));
}

static inline void write_cr3(ureg_t cr3)
{
    asm volatile("mov %0, %%cr3"
		 :: "r" (cr3)
		 : "memory");
}

static inline void write_cr4(ureg_t val)
{
    asm volatile("mov %0, %%cr4"
		 :: "r" (val));
}

static inline void flush_tlb(void)
{
    ureg_t tmpreg;

    asm volatile("mov %%cr3, %0;              \n"
		 "mov %0, %%cr3;  # flush TLB \n"
		 : "=r" (tmpreg)
		 :: "memory");
}

static inline void flush_tlb_addr(ureg_t va)
{
    asm volatile("invlpg (%0)" ::"r" (va) : "memory");
}

static inline void rdmsr(uint32_t msr, uint32_t *ax, uint32_t *dx)
{
    asm volatile("rdmsr"
		 : "=a" (*ax), "=d" (*dx)
		 : "c" (msr));
}

static inline void wrmsr(uint32_t msr, uint32_t ax, uint32_t dx)
{
    asm volatile("wrmsr"
		 : /* no outputs */
		 : "c" (msr), "a" (ax), "d" (dx));
}

static inline void wrmsrl(uint32_t msr, uint64_t val)
{
    uint32_t ax = (uint32_t)val;
    uint32_t dx = (uint32_t)(val >> 32);
    wrmsr(msr, ax, dx);
}

static inline int wrmsrl_safe(uint32_t msr, uint64_t val)
{
    uint32_t ax = (uint32_t)val;
    uint32_t dx = (uint32_t)(val >> 32);
    return wrmsr_safe(msr, ax, dx);
}

static inline void lldt(uint16_t sel)
{
    asm volatile("lldt %0"
		 : /* no outputs */
		 : "a" (sel));
}

static inline void ltr(uint16_t sel)
{
    asm volatile("ltr %0"
		 : /* no outputs */
		 : "a" (sel));
}

static inline void halt(void)
{
    vminfo.running=0;
    vminfo.blocked=1;
    asm volatile("hlt" : : );
    vminfo.running=1;
    vminfo.blocked=0;
}

static inline void clts(void)
{
    asm volatile("clts" : : );
}

static inline void vmexit_panic(void)
{
    asm volatile("outb %%al,$0xeb" : : : "memory");
}

static inline void shutdown(void)
{
    asm volatile("outb %%al,$0xed" : : : "memory");
}

